#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>

#include "vote.h"

#ifdef FALSE
#undef FALSE
#endif
#define FALSE 0

#ifdef TRUE
#undef TRUE
#endif
#define TRUE (!(FALSE))

/* This array contains the names of the involved political parties. */
char Party[__VOTE_H__NPARTIES][__VOTE_H__CHAR_BUFFER_SIZE + 1] =
{
   "Blue",
   "Yellow",
   "Green",
   "Red",
   "Purple",
   "Black",
   "White"
};

/* This function sets the (VP->Voting)[] to zero! */
/* It returns FALSE on error!                     */
int VoteZERO(t_Vote *VP)
{
   int i;

   /* Exit witth FALSE on error in input! */
   if (!VP) return(FALSE);
   if (__VOTE_H__NPARTIES <= 0) return(FALSE);

   for (i=0; i<__VOTE_H__NPARTIES; i++)
   {
      (VP->Voting)[i] = 0.0;
   }

   return(TRUE);
}

/* This functions initializes the data of *CP. */
/* It returns FALSE on error!                  */
int CitizenInit(t_Citizen *CP, char *Name)
{
   int B;
   int l;
   char *P;
   
   /* Exit with FALSE on error in input! */
   if (!CP) return(FALSE);
   if (!Name) return(FALSE);
   if (__VOTE_H__CHAR_BUFFER_SIZE <= 0) return(FALSE);
   l = strlen(Name);
   if (l < 0) return(FALSE);
   if (l > __VOTE_H__CHAR_BUFFER_SIZE) return(FALSE);

   B = VoteZERO(&(CP->Vote));
   /* Exit with FALSE on error! */
   if (!B) return(FALSE);

   P = strcpy(CP->Name, Name);
   /* Exit with FALSE on error!  */
   if (!P) return(FALSE);

   return(TRUE);
}

/* This function checks *VP for possible overflow.      */
/* This function ensures the computation of VoteSumAbs. */
/* It returns FALSE on error!                           */
int VoteSumAbs_Check(t_Vote *VP)
{
   int i;
   double sum;
   double x;

   /* Exit with FALSE on error in input! */
   if (!VP) return(FALSE);
   if (__VOTE_H__NPARTIES <= 0) return(FALSE);

   /* initialize */
   sum = 0.0;

   /* loop through the Vote *VP */
   for (i=0; i<__VOTE_H__NPARTIES; i++)
   {
      /* Fetch single Voting! */
      x = fabs((VP->Voting)[i]);
      
      /* Check for overflow! */
      if (sum > (DBL_MAX - x)) return(FALSE);

      /* register x */
      sum += x;
   }

   return(TRUE);
}

/* This function computes the sum of all Votings of VP.               */
/* You must call VoteSumAbs_Check(VP) before computing this function! */
/* (It assumes, that there is no overflow in *VP.)                    */
/* If you get an error, you just forgot to call VoteSumAbs_Check(VP). */
double VoteSumAbs(t_Vote *VP)
{
   int i;
   double sum;

   /* If you ran VoteSumAbs_Check(VP),             */
   /* you can be sure, that all data of *VP is OK! */

   /* initialize */
   sum = 0.0;

   /* loop through the Vote *VP */
   for (i=0; i<__VOTE_H__NPARTIES; i++)
   {
      /* Register single Voting! */
      sum += fabs((VP->Voting)[i]);
   }

   return(sum);
}

/* This function check *VP for possible errors. */
/* If there is such an error ite returns FALSE. */
int VoteIsZERO_Check(t_Vote *VP)
{
   int B;

   /* Exit with FALSE on error in input! */
   if (!VP) return(FALSE);

   B = VoteSumAbs_Check(VP);
   /* Exit with FALSE on error! */
   if (!B) return(FALSE);

   return(TRUE);
}

/* This function decides wether all Votings of *VP are zero!          */
/* You must call VoteIsZERO_Check(VP) before computing this function! */
/* (It assumes, that the data of *VP is OK.)                          */
/* If you get an error, you just forgot to call VoteIsZERO_Check(VP)! */
int VoteIsZERO(t_Vote *VP)
{
   double sum;

   /* If you ran VoteIsZERO_Check(VP),             */
   /* you can be sure, that all data of *VP is OK! */

   /* sum will be >= 0.0! */
   sum = VoteSumAbs(VP);
   if (sum == 0.0) return(TRUE);
   /* Here sum is > 0.0! */

   return(FALSE);
}

/* This function normalizes the data of *VP to percentage! */
/* It returns FALSE on error!                              */
int VoteNormalize100(t_Vote *VP)
{
   int B;
   int i;
   double sum;
   double x;
   double x0;
   
   /* Exit with FALSE on error in input! */
   if (!VP) return(FALSE);
   if (__VOTE_H__NPARTIES <= 0) return(FALSE);

   /* We want to call VoteIsZERO(VP)! */
   B = VoteIsZERO_Check(VP);
   /* Exit wtih FALSE on error! */
   if (!B) return(FALSE);

   /* Fast Exit with TRUE if all (VP->Voting)[] are zero! */
   if (VoteIsZERO(VP)) return(TRUE);
   /* From here it exists one i with ((VP->Voting)[i] != 0.0)! */
   /* A call sum = VoteSumAbs(VP) will result in (sum > 0.0).  */

   /* We want to call VoteSumAbs(VP)! */
   B = VoteSumAbs_Check(VP);
   /* Exit with FALSE on error! */
   if (!B) return(FALSE);

   /* Now compute sum of all fabs((VP->Voting)[]) */
   sum = VoteSumAbs(VP);
   /* Here is sum > 0.0!  */
   /* This cannot happen! */
   if (sum == 0.0) return(FALSE);
   /* Here (1.0 / sum) is allowed! */
   /* So we have (0.0 < (1.0 / sum) <= DBL_MAX)! */

   for (i=0; i<__VOTE_H__NPARTIES; i++)
   {
       x = (VP->Voting)[i];
       
       /* Consider only (x != 0.0) */
       if (x != 0.0)
       {
          /* Check for overflow! */
          /* if (fabs(x / sum) > DBL_MAX) return(FALSE); */

          /* Here x0 will be > 0.0! */
          x0 = fabs(x);
          if (x0 > 1.0)
          {
             /* Cave! */
             /* Here is (x0 > 1.0). */
             /* So (DBL_MAX / x0) is allowed! */
             if ((1.0 / sum) > (DBL_MAX / x0)) return(FALSE);
          }
          else
          {
             /* Cave! */
             /* Here is (0.0 < x0 <= 1.0), */
             /* So we have (0.0 < (x0 / sum) <= (1.0 / sum) <= DBL_MAX). */
             /* NOTHING TO DO! */
          }
          
          /* Normalize to 100 percentage! */
          (VP->Voting)[i] = x / sum;
       }
   }

   return(TRUE);
}

/* This function eliminates parties with Voting < 0 from *VP! */
/* Then *VP will be normalized by VoteNormalize100(VP).       */
/* It returns FALSE on error!                                 */
int VoteTrim_gt0(t_Vote *VP)
{
   int B;
   int i;
   int sum;

   /* Exit with FALSE on error in input! */
   if (!VP) return(FALSE);
   if (__VOTE_H__NPARTIES <= 0) return(FALSE);

   /* Kick out (Votings[] < 0.0)! */
   for (i=0;i<__VOTE_H__NPARTIES; i++)
   {
      if ((VP->Voting)[i] < 0.0)
      {
         (VP->Voting)[i] = 0.0;
      }
   }

   /* Revaluate *VP! */
   B = VoteNormalize100(VP);
   /* Exit with FALSE on error! */
   if (!B) return(FALSE);

   return(TRUE);
}

/* This function contains the core.               */
/* The vote of every citizens will be considered. */
/* The votes of all citizens must be normalized.  */
/* It returns FALSE on error!                     */
int CitizenVoteAccumulate(t_Citizen C[], int NCitizens, t_Vote *ResultP)
{
   int B;
   int i;
   int j;

   double x;

   /* Exit with FALSE on error! */
   if (!C) return(FALSE);
   if (NCitizens <= 0) return(FALSE);
   if (!ResultP) return(FALSE);
   if (__VOTE_H__NPARTIES <= 0) return(FALSE);

   /* Clear ((*ResultP).Voting)[]! */
   B = VoteZERO(ResultP);
   /* Exit with FALSE on error! */
   if (!B) return(FALSE);
   
   /* Loop through the citizens and compute the result! */
   for (i=0; i<NCitizens; i++)
   {
      /* Loop through the parties and compute their result! */
      for (j=0; j<__VOTE_H__NPARTIES; j++)
      {
         x =  ((C[i]).Vote.Voting)[j];
         /* (C[i]).Vote is normalized. */
         /* Exit with FALSE on error! */
         if (fabs(x) > 1.0) return(FALSE);
         
         /* Consider only (x != 0.0)! */
         if (x != 0.0)
         {
            /* Check for overflow! */
            /* if ((fabs(x) / NCitizens) > DBL_MAX) return(FALSE); */

            /* Here is ((0.0 < fabs(x) <= 1.0) && (NCitizens >= 1.0)). */
            /* So we have (0.0 < (fabs(x) / NCitizens) <= (1.0 / NCitizens) <= 1.0 <= DBL_MAX). */
            /* NOTHING TO DO! */

            /* Accumulate single ((C[i]).Vote.Voting)[j]! */
            (ResultP->Voting)[j] += x / NCitizens;
         }
      }
   }

   return(TRUE);
}
